'''Assetto Corsa Websocket Server Class'''

import asyncio
import configparser
import hashlib
import logging
import os
import sys
from typing import Dict
import websockets

from ac_websocket_server.constants import HOST, PORT
from ac_websocket_server.error import ACWSError
from ac_websocket_server.game import ACWSGame
from ac_websocket_server.handlers import handler


def check_acserver_file(game_directory: str = None) -> bool:

    if sys.platform == 'linux':
        file_name = f'{game_directory}/acServer'
        file_hash = 'f781ddfe02e68adfa170b28d0eccbbdc'
    else:
        file_name = f'{game_directory}/acServer.exe'
        file_hash = '357e1f1fd8451eac2d567e154f5ee537'

    try:
        with open(file_name, 'rb') as file_to_check:
            data = file_to_check.read()
            if file_hash != hashlib.md5(data).hexdigest():
                return False
    except FileNotFoundError:
        return False

    return True


def check_entry_list_file(game_directory: str = None) -> bool:

    return os.path.exists(f'{game_directory}/cfg/entry_list.ini')


def check_server_cfg_file(game_directory: str = None) -> bool:

    return os.path.exists(f'{game_directory}/cfg/server_cfg.ini')


def parse_server_cfg_file(game_directory: str = None) -> configparser.ConfigParser:

    config = configparser.ConfigParser()
    config.read(f'{game_directory}/cfg/server_cfg.ini')
    return config


class ACWSServer:
    '''Represents an Assetto Corsa WebSocket Server.

    Allows control of an Assetto Corsa server with a websockets interface.'''

    def __init__(self,
                 host: str = HOST, port: int = PORT,
                 dummy: bool = False, game: str = None) -> None:

        self.logger = logging.getLogger('ac-ws.ws-server')

        self.host = host
        self.port = port

        self.dummy = dummy

        if not game:
            self.game_directory = os.getcwd()
        else:
            self.game_directory = game

        if not check_server_cfg_file(self.game_directory):
            raise ACWSError(
                f'Missing server_cfg.ini file in {self.game_directory}')

        self.server_cfg = parse_server_cfg_file(self.game_directory)

        self.server_name = self.server_cfg['SERVER']['NAME']
        self.server_cars = self.server_cfg['SERVER']['CARS']
        self.server_udp_port = self.server_cfg['SERVER']['UDP_PORT']
        self.server_tcp_port = self.server_cfg['SERVER']['TCP_PORT']
        self.server_http_port = self.server_cfg['SERVER']['HTTP_PORT']

        if not check_entry_list_file(self.game_directory):
            raise ACWSError(
                f'Missing entry_list.ini file in {self.game_directory}')

        if not check_acserver_file(self.game_directory):
            raise ACWSError(
                f'Missing or mismatched acServer binary in {self.game_directory}')

        self.game_server = None

        self.send_queue = asyncio.Queue()

        self.stop_server: asyncio.Future = None

    async def consumer(self, message):
        if b'start_server' in message:
            self.logger.info('Received request to start game server')
            try:
                self.game = ACWSGame(
                    dummy=self.dummy,
                    game_directory=self.game_directory,
                    send_queue=self.send_queue)
                await self.game.start()
                self.logger.info('Game server started')
            except ACWSError as e:
                await self.send_queue.put(str(e))
            return
        if b'stop_server' in message:
            self.logger.info('Received request to stop game server')
            try:
                await self.game.stop()
                self.logger.info('Game server stopped')
            except ACWSError as e:
                await self.send_queue.put(str(e))
            return

        response_message = f'Received unrecognised message: {message}'
        self.logger.debug(response_message)
        await self.send_queue.put(response_message)

    async def handler(self, websocket):

        await websocket.send(
            f'Welcome to the Assetto Corsa WebSocker server running at {self.host}:{self.port}')

        await websocket.send(f'Name: {self.server_name} Cars: {self.server_cars}, Ports: {self.server_udp_port} UDP, {self.server_tcp_port} TCP, {self.server_http_port} HTTP')

        await handler(websocket, self.consumer, self.producer)

    async def producer(self):
        data = await self.send_queue.get()
        return data

    async def start(self):
        '''Start the websocket server'''

        try:

            self.logger.info(f'Starting websocket server')

            self.stop_server = asyncio.Future()

            async with websockets.serve(self.handler, self.host, self.port):
                await self.stop_server

            self.logger.info(f'Stopping websocket server')

        except KeyboardInterrupt:
            self.logger.info(f'Interupting the server')

    async def stop(self):
        '''Stop the websocket server'''

        self.stop_server.set_result()



'''Assetto Corsa Websocket handlers'''

import asyncio
import websockets


async def consumer_handler(websocket, consumer):
    async for message in websocket:
        await consumer(message)


async def handler(websocket, consumer, producer):
    '''Setup consumer and producer handlers.'''

    consumer_task = asyncio.create_task(consumer_handler(websocket, consumer))
    producer_task = asyncio.create_task(producer_handler(websocket, producer))
    done, pending = await asyncio.wait(
        [consumer_task, producer_task],
        return_when=asyncio.FIRST_COMPLETED,
    )
    for task in pending:
        task.cancel()


async def producer_handler(websocket, producer):
    while True:
        await websocket.send(await producer())



'''Assetto Corsa Game Server Class'''

import asyncio
from datetime import datetime
from genericpath import isdir
import logging
import os
import sys
import websockets

from ac_websocket_server.error import ACWSError


class ACWSGame:
    '''Represents an Assetto Corsa Server.'''

    def __init__(self,  dummy: bool = False,
                 game_directory: str = None,
                 send_queue: asyncio.Queue = None) -> None:

        self.logger = logging.getLogger('ac-ws.game-server')

        self.game_directory = game_directory

        if sys.platform == 'linux':
            self.game_executable = f'{game_directory}/acServer'
        else:
            self.game_executable = f'{game_directory}/acServer.exe'

        if dummy:
            self.game_executable = 'ac_websocket_server/dummy.py'
            self.cwd = None
        else:
            self.cwd = self.game_directory

        self.process: asyncio.subprocess.Process = None

        self.send_queue = send_queue

    async def start(self):
        '''Start the game server'''

        timestamp = datetime.now().strftime("%Y%M%d_%H%M%S")

        self.logger.info(f'Starting game server')

        os.makedirs(f'{self.game_directory}/logs/session', exist_ok=True)
        os.makedirs(f'{self.game_directory}/logs/error', exist_ok=True)

        session_file = open(
            f'{self.game_directory}/logs/session/output{timestamp}.log', 'w')
        error_file = open(
            f'{self.game_directory}/logs/error/error{timestamp}.log', 'w')

        try:
            self.process = await asyncio.create_subprocess_exec(
                self.game_executable, cwd=self.cwd,
                stdout=session_file, stderr=error_file)

            self.logger.info(f'Process pid is: {self.process.pid}')
            await self.send_queue.put(f'Assetto Corsa server started')
        except PermissionError as e:
            self.logger.error(f'Process did not start: {e}')
            await self.send_queue.put(f'Assetto Corsa server did not start')
            raise ACWSError(e)

    async def stop(self):
        '''Stop the game server'''

        self.logger.info(f'Stopping game server')
        await self.send_queue.put(f'Assetto Corsa server is stopping')

        self.process.terminate()

        status_code = await asyncio.wait_for(self.process.wait(), None)
        self.logger.info(f'Game server exited with {status_code}')



#!/usr/bin/env python3

'''Dummy Game Server for development'''

import sys
import time

print('Dummy Game Server starting')
i = 0
try:
    while True:
        if i % 30 == 0:
            print(
                f'Dummy Server - Session timestamp {i} seconds')
        if i % 30 == 0:
            print(
                f'Dummy Server - Error timestamp {i} seconds', file=sys.stderr)
        time.sleep(1)
        i += 1
except KeyboardInterrupt:
    print('Dummy Game Server interrupted!')



'''Assetto Corsa Websocket Server'''

import logging
import sys


# logging.basicConfig()

# logging.basicConfig(level=logging.DEBUG,
#                     format='%(asctime)s %(levelname)s %(message)s',
#                     filename='tmp/ac-ws.log',
#                     filemode='w')

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(message)s')

logger = logging.getLogger('ac-ws')
